from sqlalchemy import *
from migrate import *
from dictionaries.versions import views_helper
import migrate.changeset


class Dictionaries(object):
    def __init__(self, meta):
        self.barcodes = Table('barcodes', meta,
                              Column('code', String(100), nullable=False),
                              Column('barcode', String(100), nullable=False, primary_key=True),
                              Column('name', String(200), nullable=True),
                              Column('price', Numeric(precision=15, scale=2), nullable=True),
                              Column('cquant', Numeric(precision=13, scale=3), nullable=True),
                              Column('measure', Integer, nullable=True),
                              )

        self.prices = Table('prices', meta,
                            Column('barcode', String(100), nullable=False),
                            Column('documentid', String(100), nullable=False),
                            Column('effectivedate', DateTime, nullable=False),
                            Column('doctype', Integer, nullable=False),
                            Column('price', Numeric(precision=15, scale=2), nullable=False),
                            )


def upgrade(migrate_engine):
    meta = MetaData(bind=migrate_engine)
    dictionaries = Dictionaries(meta)
    for table in filter(lambda x: isinstance(x, Table), map(lambda x: getattr(dictionaries, x), dir(dictionaries))):
        basename = table.name
        for suffix in ['1', '2']:
            table.name = basename + suffix
            if not table.exists(meta.bind):
                table.create()
    conn = migrate_engine.connect()
    for table in filter(lambda x: isinstance(x, Table), map(lambda x: getattr(dictionaries, x), dir(dictionaries))):
        basename = table.name[:-1]
        views_helper.create_or_replace(migrate_engine, basename, basename + '1')


def downgrade(migrate_engine):
    meta = MetaData(bind=migrate_engine)
    dictionaries = Dictionaries(meta)
    for table in filter(lambda x: isinstance(x, Table), map(lambda x: getattr(dictionaries, x), dir(dictionaries))):
        basename = table.name
        for suffix in ['1', '2']:
            table.name = basename + suffix
            if table.exists(meta.bind):
                table.drop()
    conn = migrate_engine.connect()
    for table in filter(lambda x: isinstance(x, Table), map(lambda x: getattr(dictionaries, x), dir(dictionaries))):
        basename = table.name[:-1]
        views_helper.drop_if_exist(migrate_engine, basename)
