from sqlalchemy import *
from migrate import *
import migrate.changeset

def getDepartment(meta):
    department = Table('department', meta,
        Column('departmentid', Integer, nullable=False, primary_key=True, autoincrement=True),
        Column('departmentcode', Integer, nullable=True),
        Column('departmentname', String(100), nullable=True),
        mysql_engine='InnoDB'
    )
    return department

def getDocumentTable(meta):
    return Table('document', meta, autoload=True)

def getGoodsitemTable(meta):
    return Table('goodsitem', meta, autoload=True)

def getStornogoodsitemTable(meta):
    return Table('stornogoodsitem', meta, autoload=True)


def upgrade(migrate_engine):
    meta = MetaData(bind = migrate_engine)
    department = getDepartment(meta)
    department.create()
    Column('departmentid', Integer, nullable=True).create(getDocumentTable(meta))
    Column('departmentid', Integer, nullable=True).create(getGoodsitemTable(meta))
    Column('departmentid', Integer, nullable=True).create(getStornogoodsitemTable(meta))
    migrate.changeset.constraint.ForeignKeyConstraint([getDocumentTable(meta).c.departmentid], [department.c.departmentid], name = 'documentdepartmentidfk').create()
    migrate.changeset.constraint.ForeignKeyConstraint([getGoodsitemTable(meta).c.departmentid], [department.c.departmentid], name = 'goodsitemdepartmentidfk').create()
    migrate.changeset.constraint.ForeignKeyConstraint([getStornogoodsitemTable(meta).c.departmentid], [department.c.departmentid], name = 'stornogoodsitemdepartmentidfk').create()

def downgrade(migrate_engine):
    meta = MetaData(bind = migrate_engine)
    department = getDepartment(meta)
    migrate.changeset.constraint.ForeignKeyConstraint([getDocumentTable(meta).c.departmentid], [department.c.departmentid], name = 'documentdepartmentidfk').drop()
    migrate.changeset.constraint.ForeignKeyConstraint([getGoodsitemTable(meta).c.departmentid], [department.c.departmentid], name = 'goodsitemdepartmentidfk').drop()
    migrate.changeset.constraint.ForeignKeyConstraint([getStornogoodsitemTable(meta).c.departmentid], [department.c.departmentid], name = 'stornogoodsitemdepartmentidfk').drop()
    getDocumentTable(meta).c.departmentid.drop()
    getGoodsitemTable(meta).c.departmentid.drop()
    getStornogoodsitemTable(meta).c.departmentid.drop()
    department.drop()
