# -*- coding: utf-8 -*-

from sqlalchemy import MetaData
import os
import time
import pymysql as mysql
import datetime
import re
from documents.versions import conf
from artix.database.info import DBInfo

meta = MetaData()

workDir = '/linuxcash/cash/data/db/tmp/'
errorDir = '/linuxcash/logs/defects/database/'
flagBlockFile = os.path.join(workDir, 'backup_data.lck')
flagRestoreFile = os.path.join(workDir, 'restore_data.lck')
ignoreList = ['documenttype', 'goodsitemtype', 'migrate_version', 'moneyitemtype', 'shmdiscounts', 
              'shmdiscountsdetail', 'shmdiscountsdetailext', 'shmdiscountsext', 'shmmoneyoperations', 'shmpositions', 
              'shmpositionsstorno', 'shmstorno', 'text_const']

[os.makedirs(x) for x in [workDir, errorDir] if not os.path.exists(x)]

def logError(error):
    with open(os.path.join(errorDir, datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S.log")), 'a') as f:
        f.write('\n%s' % error)

def removeFile(fileName):

    try:
        os.remove(fileName)
    except OSError:
        pass

def isExistTable(migrate_engine, tableName):

    a = migrate_engine.execute("show tables like '%s';" % tableName)

    if hasattr(a, 'fetchone'):
        return a.fetchone()

    return a

def needToStore(cursor):

    result = cursor.execute("SELECT COUNT(*) FROM documents.workshift;").fetchone()
    resultGoods = cursor.execute("SELECT COUNT(*) FROM documents.goodsitem;").fetchone()
    return result and int(result[0]) > conf.shiftCount and resultGoods and int(resultGoods[0]) > conf.goodsItemCount

def getShiftId(cursor):

    if not isExistTable(cursor, 'backup_workshift'):
        return  ['workshiftid', []]

    cursor.execute('select workshiftid from documents.backup_workshift order by workshiftid desc limit 3;')
    return  ['workshiftid', [str(k[0]) for k in cursor.fetchall()]]

def getDocumentId(cursor, shiftIds):

    if not shiftIds:
        return ['documentid', []], ['buttonid', []], ['departmentid', []]

    query = 'select documentid, buttonid, departmentid from documents.backup_document where workshiftid in (%s);' % ','.join(shiftIds)

    listDocId = []
    listButtonId = []
    listDepartmentId = []

    cursor.execute(query)
    for ids in cursor.fetchall():
        listDocId.append(str(ids[0]))
        if ids[1]:
            listButtonId.append(str(ids[1]))
        if ids[2]:
            listDepartmentId.append(str(ids[2]))
    return ['documentid', listDocId], ['buttonid', listButtonId], ['departmentid', listDepartmentId]

def getGoodsItemId(cursor, documentIds):

    if not documentIds:
        return ['goodsitemid', []]

    query = 'select goodsitemid from documents.backup_goodsitem where documentid in (%s);' % ','.join(documentIds)
    listId = []
    cursor.execute(query)
    for goodsItemId in cursor.fetchall():
        listId.append(str(goodsItemId[0]))
    return ['goodsitemid', listId]

def getStornoGoodsItemId(cursor, documentIds):

    if not documentIds:
        return ['stornogoodsitemid', []]

    query = 'select stornogoodsitemid from documents.backup_stornogoodsitem where documentid in (%s);' % ','.join(documentIds)
    listId = []
    cursor.execute(query)
    for goodsItemId in cursor.fetchall():
        listId.append(str(goodsItemId[0]))
    return ['stornogoodsitemid', listId]

def getCardItemId(cursor, documentIds):

    if not documentIds:
        return  ['carditemid', []], ['clientitemid', []]

    query = 'SELECT carditemid, clientitemid FROM documents.backup_carditem where documentid in (%s);' % ','.join(documentIds)
    listCardId = []
    listClientId = []
    cursor.execute(query)
    for ids in cursor.fetchall():
        listCardId.append(str(ids[0]))
        if ids[1]:
            listClientId.append(str(ids[1]))
    return  ['carditemid', listCardId], ['clientitemid', listClientId]

def getPaymentId(cursor, goodsItemIds, stornoGoodsItemIds):

    listPaymentId = []

    if goodsItemIds:
        query = 'SELECT distinct(paymentitemid) FROM documents.backup_goodsitem where goodsitemid in (%s) and paymentitemid IS NOT NULL;' % ','.join(goodsItemIds)
        cursor.execute(query)
        for paymentId in cursor.fetchall():
            listPaymentId.append(str(paymentId[0]))

    if stornoGoodsItemIds:
        query = 'SELECT distinct(paymentitemid) FROM documents.backup_stornogoodsitem where stornogoodsitemid in (%s) and paymentitemid IS NOT NULL;' % ','.join(stornoGoodsItemIds)
        cursor.execute(query)
        for paymentId in cursor.fetchall():
            listPaymentId.append(str(paymentId[0]))

    return ['paymentitemid', listPaymentId]

def getDocTables(migrate_engine):

    # Важен порядок при бэкапе. 
    # Сначала должна быть смена, затем документ и т.д.
    # Иначе в таблицах будут некорректные foreign keys
    tables = ['workshift','kkm','crights','users',
                  'client','button','paymentitem','department','document',
                  'carditem','couponitem','moneyitem','goodsitem','stornogoodsitem',
                  'certificate', 'bonusitem','aspectitem','discitem','alcosetitem',
                  'cashitem', 'securityweight']

    # Добавим остальные таблицы в наш список
    result = migrate_engine.execute("show tables from documents where Tables_in_documents not LIKE 'backup_%%';")
    t = migrate_engine.fetchall() if hasattr(migrate_engine, 'fetchall') else result.fetchall()
    for table in t:
        if table[0] not in ignoreList and table[0] not in tables:
            tables.append(table[0])
    return tables

def renameTables(migrate_engine):
    migrate_engine.execute('SET autocommit = 0;')
    migrate_engine.execute('SET foreign_key_checks = 0;')

    for tableName in getDocTables(migrate_engine):
        if isExistTable(migrate_engine, tableName) and not isExistTable(migrate_engine, 'backup_%s' % tableName):

            migrate_engine.execute('RENAME TABLE documents.%s TO documents.backup_%s;' % (tableName, tableName))
            q = migrate_engine.execute('SHOW CREATE TABLE documents.backup_%s;' % (tableName)).fetchone()[1]

            q = q.replace('backup_', '')
            substrings = re.findall(r'CONSTRAINT `([^E]+)` FOREIGN', q)
            for k in substrings:

                constraint_name = k.replace("_1", "") if "_1" in k else "%s_1" % k

                key_old = "CONSTRAINT `%s` FOREIGN" % k
                key_new = "CONSTRAINT `%s` FOREIGN" % constraint_name

                q = q.replace(key_old, key_new)
            migrate_engine.execute(q)

    dropTriggers(migrate_engine)
    createTriggers(migrate_engine)

def store(migrate_engine):
    if not conf.enable:
        return

    if os.path.isfile(flagBlockFile):
        return

    meta.bind = migrate_engine

    if 'documents_development' in str(migrate_engine.url):
        return

    if not needToStore(migrate_engine):
        return

    print('')
    print("store data...")

    os.system('service nes stop')
    os.system('service exchangers-unload stop')
    os.system('service artix-aes-shtrihm-unload stop')

    runningTime = time.time()

    renameTables(migrate_engine)
    open(flagBlockFile, 'a').close()

    print("full time store: %s" % format(time.time() - runningTime, "0.5f"))


def getLastId(cursor, tableName, tableId):
    '''
    Получить последний идентификатор из таблицы
    '''
    if not tableId:
        return

    q = "SELECT MAX(%s) as MaximumID FROM documents.%s;" % (tableId, tableName)

    resExecute = cursor.execute(q)

    result = cursor.fetchone()[0] if resExecute else None

    if result:
        return str(result)

    return None

def restoreTable(cursor, listId, tableName, tableId):
    """
    Восстановим данные из таблиц backup_*
    listId - список идентификаторов по которым нужно восстановить данные
    tableName - имя таблицы
    tableId - идентификатор таблицы.
    """

    idName = listId[0]
    ids = listId[1]

    if not isExistTable(cursor, 'backup_%s' % tableName):
        return

    if not ids:     # В последних 3-х сменах не обнаружены данные.
                    # Найдем, есть ли вообще записи

        if not tableId:
            return

        lastId = getLastId(cursor, 'backup_%s' % tableName, tableId)

        if not lastId:
            return

        idName = tableId
        ids = [lastId]


    listColumn = []
    q = "SELECT column_name FROM information_schema.columns WHERE table_name = 'backup_%s' and table_schema = 'documents' ORDER  BY ordinal_position;" % tableName
    cursor.execute(q)
    for a in cursor.fetchall():
        listColumn.append("`%s`" % (a[0]))

    q = "INSERT INTO documents.%s (%s) select %s from documents.backup_%s g where g.%s in (%s);" % (tableName, ','.join(listColumn), ','.join(listColumn), tableName, idName, ','.join(ids))
    cursor.execute(q)

    q = "DELETE FROM documents.backup_%s WHERE %s in (%s);" % (tableName, idName, ','.join(ids))
    cursor.execute(q)

    # Проверим есть ли записи в таблице, если нет восстановим одну из backup
    q = "SELECT 1 FROM documents.%s LIMIT 1;" % tableName
    res = cursor.execute(q)
    if not res:
        restoreTable(cursor, [0, []], tableName, tableId)

def restore():

    if not os.path.isfile(flagBlockFile):
        return

    print('')
    print('restore data...')

    t0 = time.time()
    dbInfo = DBInfo()
    db = mysql.connect(host=dbInfo.host, port=dbInfo.port, user=dbInfo.user, password=dbInfo.password, db="documents")
    try:
        db.autocommit(False)
        cur = db.cursor()
        cur.execute('SET foreign_key_checks = 0;')
        cur.execute('SET @TRIGGER_AFTER_INSERT = FALSE;')
        cur.execute('START TRANSACTION;')

        shiftIds = getShiftId(cur)
        documentIds, buttonIds, departmentIds = getDocumentId(cur, shiftIds[1])
        goodsItemIds = getGoodsItemId(cur, documentIds[1])
        stornoGoodsItemIds = getStornoGoodsItemId(cur, documentIds[1])
        cardIds, clientIds = getCardItemId(cur, documentIds[1])
        paymentIds = getPaymentId(cur, goodsItemIds[1], stornoGoodsItemIds[1])

        restoreTable(cur, paymentIds, 'paymentitem', 'paymentitemid')
        restoreTable(cur, cardIds, 'couponitem', 'couponid')
        restoreTable(cur, clientIds, 'client', 'clientitemid')
        restoreTable(cur, departmentIds, 'department', 'departmentid')
        restoreTable(cur, buttonIds, 'button', 'buttonid')
        restoreTable(cur, goodsItemIds, 'discitem', 'discitemid')
        restoreTable(cur, goodsItemIds, 'aspectitem', 'aspectitemid')
        restoreTable(cur, goodsItemIds, 'alcosetitem', 'alcosetitemid')
        restoreTable(cur, goodsItemIds, 'securityweight', 'securityweightid')
        restoreTable(cur, documentIds, 'certificate', 'certificateid')
        restoreTable(cur, documentIds, 'stornogoodsitem', 'stornogoodsitemid')
        restoreTable(cur, documentIds, 'moneyitem', 'moneyitemid')
        restoreTable(cur, documentIds, 'goodsitem', 'goodsitemid')
        restoreTable(cur, documentIds, 'carditem', 'carditemid')
        restoreTable(cur, documentIds, 'bonusitem', 'bonusitemid')
        restoreTable(cur, documentIds, 'cashitem', 'cashitemid')
        restoreTable(cur, shiftIds, 'crights', 'crightsid')
        restoreTable(cur, shiftIds, 'users', 'workshiftid')
        restoreTable(cur, shiftIds, 'kkm', 'kkmid')
        restoreTable(cur, shiftIds, 'document', 'documentid')
        restoreTable(cur, shiftIds, 'workshift', None)

        db.commit()
        cur.close()
        db.close()
    except Exception as e:
        logError(str(e))
        db.rollback()
        db.close()
        raise

    print("full time restore: %s" % format(time.time() - t0, "0.5f"))
    removeFile(flagBlockFile)
    open(flagRestoreFile, 'a').close()

    os.system('service nes start --no-wait')
    os.system('service exchangers-unload start --no-wait')

def getTrAfterInsert():
    dbInfo = DBInfo()
    return """
CREATE TRIGGER `triggerAfterInsert` AFTER INSERT ON `document`
FOR EACH ROW
thisTrigger: BEGIN
    if (@TRIGGER_AFTER_INSERT = FALSE) AND (USER() = '%s@%s') THEN
        LEAVE thisTrigger;
    END IF;
    SET @factor:=(SELECT moneyfactor FROM documenttype WHERE doctype = new.doctype);
    UPDATE workshift
    SET sumsale = COALESCE(sumsale, 0) + if(new.doctype = 1, new.sumb, 0),
        sumsalecash = COALESCE(sumsalecash, 0) + if(new.doctype = 1, new.sumcash, 0),
        sumsalenoncash = COALESCE(sumsalenoncash, 0) + if(new.doctype = 1, new.sumnoncash, 0),
        sumsaleother = COALESCE(sumsalenoncash, 0) + if(new.doctype = 1, new.sumother, 0),
        sumgaincash = COALESCE(sumgaincash, 0) + if(new.doctype in (1,2,25), new.sumcash * @factor, 0),
        sumgainnoncash = COALESCE(sumgainnoncash, 0) + if(new.doctype in (1,2,25), new.sumnoncash * @factor, 0),
        sumrefund = COALESCE(sumrefund, 0) + if(new.doctype in (2,25), abs(new.sumb), 0),
        sumrefundcash = COALESCE(sumrefundcash, 0) + if(new.doctype in (2,25), new.sumcash, 0),
        sumrefundnoncash = COALESCE(sumrefundnoncash, 0) + if(new.doctype in (2,25), new.sumnoncash, 0),
        sumgain = COALESCE(sumgain, 0) + if(new.doctype in (1, 2, 25), new.sumb, 0),
        sumdrawer = COALESCE(sumdrawer, 0) + (new.sumcash * @factor),
        countsale = COALESCE(countsale, 0) + if(new.doctype = 1, 1, 0),
        countrefund = COALESCE(countrefund, 0) + if(new.doctype in (2, 25), 1, 0)
    WHERE workshiftid = new.workshiftid AND new.closed = 1;
END
""" % (dbInfo.user, dbInfo.host)

def getTrAfterUpdate():
    return """
CREATE TRIGGER `triggerAfterUpdate` AFTER UPDATE ON `document`
FOR EACH ROW BEGIN
    set @oldfactor:=(select moneyfactor FROM documenttype WHERE doctype = old.doctype);
    set @newfactor:=(select moneyfactor FROM documenttype WHERE doctype = new.doctype);
    set @oldsumb=0;
    set @oldsumcash=0;    
    set @newsumb=0;    
    set @newsumcash=0;
    set @oldsumnoncash=0;
    set @newsumnoncash=0;
    set @oldcount=0;
    set @newcount=0;
    set @oldsumother=0;
    set @newsumother=0;

    IF old.closed = 1
    THEN
       set @oldsumb=old.sumb;
       set @oldsumother=old.sumother;
       set @oldsumcash=@oldfactor * old.sumcash;
       set @oldsumnoncash=@oldfactor * old.sumnoncash;
       set @oldcount=1;
    END IF;

    IF new.closed = 1
    THEN
       set @newsumb=new.sumb;
       set @newsumother=new.sumother;
       set @newsumcash=@newfactor * new.sumcash;
       set @newsumnoncash=@newfactor * new.sumnoncash;
       set @newcount=1;
    END IF;

    UPDATE workshift
    SET sumsale = COALESCE(sumsale, 0) + if(new.doctype = 1, -@oldsumb + @newsumb, 0),
        sumgain = COALESCE(sumgain, 0) + if(new.doctype in (1, 2, 25), -@oldsumb + @newsumb, 0),
        sumdrawer = COALESCE(sumdrawer, 0) - @oldsumcash + @newsumcash,        
        sumrefund = COALESCE(sumrefund, 0) + if(new.doctype in (2,25), -abs(@oldsumb) + abs(@newsumb), 0),
        sumsalecash = COALESCE(sumsalecash, 0) + if(new.doctype = 1, -@oldsumcash + @newsumcash, 0),        
        sumsalenoncash = COALESCE(sumsalenoncash, 0) + if(new.doctype = 1, -@oldsumnoncash + @newsumnoncash, 0),        
        sumsaleother = COALESCE(sumsaleother, 0) + if(new.doctype = 1, -@oldsumother + @newsumother, 0),        
        sumgaincash = COALESCE(sumgaincash, 0) + if(new.doctype in (1,2,25), -@oldsumcash + @newsumcash, 0),        
        sumgainnoncash = COALESCE(sumgainnoncash, 0) + if(new.doctype in (1,2,25), -@oldsumnoncash + @newsumnoncash, 0),        
        sumrefundcash = COALESCE(sumrefundcash, 0) + if(new.doctype in (2,25), -abs(@oldsumcash) + abs(@newsumcash), 0),        
        sumrefundnoncash = COALESCE(sumrefundnoncash, 0) + if(new.doctype in (2,25), -abs(@oldsumnoncash) + abs(@newsumnoncash), 0),        
        countsale = COALESCE(countsale, 0) + if(new.doctype = 1, -@oldcount + @newcount, 0),
        countrefund = COALESCE(countrefund, 0) + if(new.doctype in (2, 25), -@oldcount + @newcount, 0)
    WHERE workshiftid = new.workshiftid;
END"""

def getTrAfterDelete():
    return """
CREATE TRIGGER `triggerAfterDelete` AFTER DELETE ON `document`
FOR EACH ROW BEGIN
    SET @factor:=(SELECT moneyfactor FROM documenttype WHERE doctype = old.doctype);
    UPDATE workshift
    SET sumsale = COALESCE(sumsale, 0) - if(old.doctype = 1, old.sumb, 0),
        sumgain = COALESCE(sumgain, 0) - if(old.doctype in (1, 2, 25), old.sumb, 0),
        sumdrawer = COALESCE(sumdrawer, 0) - (old.sumcash * @factor),
        sumrefund = COALESCE(sumrefund, 0) - if(old.doctype in (2,25), abs(old.sumb), 0),        
        sumsalecash = COALESCE(sumsalecash, 0) - if(old.doctype = 1, old.sumcash, 0),                
        sumsalenoncash = COALESCE(sumsalenoncash, 0) - if(old.doctype = 1, old.sumnoncash, 0),              
        sumsaleother = COALESCE(sumsaleother, 0) - if(old.doctype = 1, old.sumother, 0),                
        sumgaincash = COALESCE(sumgaincash, 0) - if(old.doctype in (1,2,25), old.sumcash * @factor, 0),               
        sumgainnoncash = COALESCE(sumgainnoncash, 0) - if(old.doctype in (1,2,25), old.sumnoncash * @factor, 0),                
        sumrefundcash = COALESCE(sumrefundcash, 0) - if(old.doctype in (2,25), old.sumcash, 0),               
        sumrefundnoncash = COALESCE(sumrefundnoncash, 0) - if(old.doctype in (2,25), old.sumnoncash, 0),             
        countsale = COALESCE(countsale, 0) - if(old.doctype = 1, 1, 0),
        countrefund = COALESCE(countrefund, 0) - if(old.doctype in (2, 25), 1, 0)
    WHERE workshiftid = old.workshiftid AND old.closed = 1;
END"""

def getOldTrAfterInsert(docTypes):
    dbInfo = DBInfo()
    return """
CREATE TRIGGER `triggerAfterInsert` AFTER INSERT ON `document`
FOR EACH ROW
thisTrigger: BEGIN
    if (@TRIGGER_AFTER_INSERT = FALSE) AND (USER() = '%s@%s') THEN
        LEAVE thisTrigger;
    END IF;
    SET @factor:=(SELECT moneyfactor FROM documenttype WHERE doctype = new.doctype);
    UPDATE workshift
    SET sumsale = COALESCE(sumsale, 0) + IF(new.doctype = 1, new.sumb, 0),
        sumgain = COALESCE(sumgain, 0) + IF(new.doctype IN %s, new.sumb, 0),
        sumdrawer = COALESCE(sumdrawer, 0) + (new.sumcash * @factor)
    WHERE workshiftid = new.workshiftid AND new.closed = 1;
END
""" % (dbInfo.user, dbInfo.host, docTypes)

def getOldTrAfterUpdate(docTypes):
    return """
CREATE TRIGGER `triggerAfterUpdate` AFTER UPDATE ON `document`
FOR EACH ROW BEGIN
    SET @oldfactor:=(SELECT moneyfactor FROM documenttype WHERE doctype = old.doctype);
    SET @newfactor:=(SELECT moneyfactor FROM documenttype WHERE doctype = new.doctype);
    SET @oldsumb=0;
    SET @oldsumcash=0;
    SET @newsumb=0;
    SET @newsumcash=0;

    IF old.closed = 1
    THEN
       SET @oldsumb=old.sumb;
       SET @oldsumcash=@oldfactor * old.sumcash;
    END IF;

    IF new.closed = 1
    THEN
       SET @newsumb=new.sumb;
       SET @newsumcash=@newfactor * new.sumcash;
    END IF;

    UPDATE workshift
    SET sumsale = COALESCE(sumsale, 0) + IF(new.doctype = 1, -@oldsumb + @newsumb, 0),
        sumgain = COALESCE(sumgain, 0) + IF(new.doctype IN %s, -@oldsumb + @newsumb, 0),
        sumdrawer = COALESCE(sumdrawer, 0) - @oldsumcash + @newsumcash
    WHERE workshiftid = new.workshiftid;
END""" % docTypes

def getOldTrAfterDelete(docTypes):
    return """
CREATE TRIGGER `triggerAfterDelete` AFTER DELETE ON `document`
FOR EACH ROW BEGIN
    SET @factor:=(SELECT moneyfactor FROM documenttype WHERE doctype = old.doctype);
    UPDATE workshift
    SET sumsale = COALESCE(sumsale, 0) - IF(old.doctype = 1, old.sumb, 0),
        sumgain = COALESCE(sumgain, 0) - IF(old.doctype IN %s, old.sumb, 0),
        sumdrawer = COALESCE(sumdrawer, 0) - (old.sumcash * @factor)
    WHERE workshiftid = old.workshiftid AND old.closed = 1;
END""" % docTypes


def dropTriggers(migrate_engine):
    migrate_engine.connect().execute("DROP TRIGGER IF EXISTS `triggerAfterInsert`;")
    migrate_engine.connect().execute("DROP TRIGGER IF EXISTS `triggerAfterUpdate`;")
    migrate_engine.connect().execute("DROP TRIGGER IF EXISTS `triggerAfterDelete`;")

def createTriggers(migrate_engine, docTypes = "(1, 2, 25)"):
    version  = migrate_engine.execute("SELECT version FROM documents.migrate_version;").fetchone()
    if int(version[0]) > 141:
        migrate_engine.execute(getTrAfterInsert())
        migrate_engine.execute(getTrAfterUpdate())
        migrate_engine.execute(getTrAfterDelete())
    else:
        migrate_engine.execute(getOldTrAfterInsert(docTypes))
        migrate_engine.execute(getOldTrAfterUpdate(docTypes))
        migrate_engine.execute(getOldTrAfterDelete(docTypes))
